#include <iostream>
#include <ostream>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stddef.h>
#include "board.h"
#include "assert.h"

namespace gol
{

    /** Constructor for a game board
     *
     * Initialise the grid, fill it with dead cells.
     * Create the history array if required.
     * Add a glider.
     */
    Board::Board(uint8_t width, uint8_t height, uint16_t historyLevels) :
        width(width), height(height), historyLevels(historyLevels)
    {
        assert(width > 0);
        assert(height > 0);

        this->grid = new uint8_t[width * height];
        memset(this->grid, 0, sizeof(uint8_t) * width * height);

        if (historyLevels > 0)
        {
            this->history = new uint8_t*[historyLevels];
            for (int i = 0; i < historyLevels; ++i)
                this->history[i] = NULL;
        }
        else
        {
            this->history = NULL;
        }

        putGlider();
    }


    /** Clean up the game board */
    Board::~Board()
    {
        if (NULL != this->history)
        {
            for (uint16_t h = 0; h < this->historyLevels; ++h)
            {
                delete[] this->history[h];
            }
            delete[] this->history;
        }
        delete[] this->grid;
    }


    /** Draw a glider at 0, 0
     */
    void Board::putGlider(void)
    {
        if ((this->width >= 3) && (this->height >= 3))
        {
            this->grid[cell(1,0)] = this->grid[cell(2,1)] = this->grid[cell(0,2)] = this->grid[cell(1,2)] = this->grid[cell(2,2)] = 1;
        }
    }


    /** Fill the board with a random assortment of alive/dead cells.
     */
    void Board::randomiseBoard(uint8_t fillFactor)
    {
        for (int x = 0; x < this->width; ++x)
        {
            for (int y = 0; y < this->width; ++y)
            {
                uint8_t r = (uint8_t)rand();
                this->grid[cell(x, y)] = (r < fillFactor);
            }
        }
    }


    /** Get the board width. */
    uint8_t Board::getWidth() const
    {
        return this->width;
    }

    
    /** Get the board height. */
    uint8_t Board::getHeight() const
    {
        return this->height;
    }

    /** Get the array offset in grid[] for a given cell. */
    uint16_t Board::cell(uint8_t x, uint8_t y) const
    {
        assert(x < this->width);
        assert(y < this->height);

        return (y * this->width) + x;
    }


    /** Display the board on an output stream.
     */
    std::ostream& operator<<(std::ostream &os, const Board &b)
    {
        os << "Game of Life Board (" << (int)b.getWidth() << "x" << (int)b.getHeight() << ")" << std::endl;
        for (int y = 0; y < b.height; ++y)
        {
            for (int x = 0; x < b.width; ++x)
            {
                os << (b.grid[b.cell(x,y)] ? "X " : ". ");
            }

            os << std::endl;
        }
    }


    /** Record the current state in the game history
     */
    void Board::recordState(void)
    {
        if (this->historyLevels < 1)
            return;

        assert(NULL != this->history);
        for (int i = 1; i < this->historyLevels; ++i)
        {
            this->history[i-1] = this->history[i];
        }

        this->history[this->historyLevels - 1] = new uint8_t[this->width * this->height];
        memcpy(this->history[this->historyLevels - 1], this->grid, sizeof(uint8_t) * this->width * this->height);
    }


    /** Check to see whether the current game state is recorded in the history. */
    int Board::checkStateMatch(void) const
    {
        assert((this->historyLevels == 0) || (this->history != NULL));

        for (int state = this->historyLevels - 1; state >= 0; --state)
        {
            if (NULL == this->history[state])
                continue;

            if (memcmp(this->grid, this->history[state], sizeof(uint8_t) * this->width * this->height) == 0)
            {
                return state;
            }
        }

        return -1;
    }
}
