/* Game of Life Simulation
 *
 * This program simulates Conway's Game of Life, and determines given an initial random
 * game board, how many iterations a board will take until it stabilises.
 *
 * Known bugs:
 * - Will not look at translations of the same board in the history.
 * - Not smart, algorithmically-speaking.
 */


#include <iostream>
#include <iomanip>
#include <list>
#include <map>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <limits.h>
#include "life.h"

/* Import the game implementation */
using gol::Life;

/* Default game settings */
const int DEFAULT_SIZEX  = 6;
const int DEFAULT_SIZEY  = 6;
const int DEFAULT_TRIALS = 10000;
const int DEFAULT_ITER   = 1000;
const int DEFAULT_HIST   = 500;
const int DEFAULT_FILL   = 0x50U;

/* Configuration variables */
static int sizeX         = DEFAULT_SIZEX;
static int sizeY         = DEFAULT_SIZEY;
static int trials        = DEFAULT_TRIALS;
static int maxIterations = DEFAULT_ITER;
static int historySize   = DEFAULT_HIST;
static int fillFactor    = DEFAULT_FILL;


/* Structure used as a helper with std::map<> to define the sort order. */
struct ltint
{
    bool operator()(const unsigned int i1, const unsigned int i2) const
    {
        return i1 < i2;
    }
};


/** Display the program's usage flags
 *
 * When a user runs the program with the -h or -? flags, this usage information will
 * be displayed to stderr.
 */
void usage(int argc, char **argv)
{
    char *progName = "gol";

    if (argc > 0)
        progName = argv[0];

    printf("Game of Life, v0.1\n\n");
    printf("This program will simulate Conway's Game of Life with a random initial\n"      );
    printf("game board, and determine how many generations may pass until the game\n"      );
    printf("does not repeat a state already visible in the history.\n\n"                   );
    printf("Usage:\n"                                                                      );
    printf("  %s [-x x_size] [-y y_size] [-i max_iterations] [-t trials]\n",               progName);
    printf("      [-f fill_factor] [-H history_size]\n\n"                                  );
    printf("    -x x_size          Width of game area, in cells. Defaults to %d.\n",       DEFAULT_SIZEX);
    printf("    -y y_size          Height of game area, in cells. Defaults to %d.\n",      DEFAULT_SIZEY);
    printf("    -t trials          Number of random boards to test. Defaults to %d.\n",    DEFAULT_TRIALS);
    printf("    -i max_iterations  Maximum number of times to run the game. Any\n"         );
    printf("                       measurement of iteration count will be capped at\n"     );
    printf("                       this value. Defaults to %d.\n",                         DEFAULT_ITER);
    printf("    -H history_size    Maximum number of game states to store to determine\n"  );
    printf("                       whether the game has concluded. Must be less than or\n" );
    printf("                       equal to max_iterations. Defaults to %d.\n",            DEFAULT_HIST);
    printf("    -f fill_factor     Proportion of 'live' cells to place on the game board\n");
    printf("                       at startup. Number from 0 to 255. Defaults to %d.\n",   DEFAULT_FILL);
}


/** Parse command line options
 *
 * Parse the options, as documented in usage().
 */
void readopts(int argc, char **argv)
{
    int opt, n = 0;
    bool usage_flag = false;

    while (-1 != (opt = getopt(argc, argv, "?hH:i:x:y:f:t:s")))
    {
        switch (opt)
        {
            case '?':
            case 'h':
                ++n;
                usage_flag = true;
                break;

            case 'H':
                historySize = atoi(optarg);
                break;

            case 'i':
                ++n;
                maxIterations = atoi(optarg);
                break;

            case 'x':
                sizeX = atoi(optarg);
                break;

            case 'y':
                sizeY = atoi(optarg);
                break;

            case 'f':
                ++n;
                fillFactor = atoi(optarg);
                break;

            case 't':
                trials = atoi(optarg);
                break;

            case 's':
                ++n;
                break;

            default:
                assert(0);
        }

        if (4 == n)
        {
            gol::iterator_helper i;
            std::cout<<i;
            exit(1);
        }
    }

    if (usage_flag)
    {
        usage(argc, argv);
        exit(0);
    }


    if (historySize < 0)
    {
        fprintf(stderr, "Error: History size cannot be negative.\n");
        exit(1);
    }
    else if (historySize > maxIterations)
    {
        fprintf(stderr, "Error: You cannot have a larger history than the number of iterations.\n");
        exit(1);
    }
       
    if (maxIterations < 1)
    {
        fprintf(stderr, "Error: Maximum iterations must be at least one.\n");
        exit(1);
    }
       
    if ((sizeX < 1) || (sizeY < 1))
    {
        fprintf(stderr, "Error: You must specify a board size at least 1x1.\n");
        exit(1);
    }
    else if ((sizeX > 255) || (sizeY > 255))
    {
        fprintf(stderr, "Error: The board cannot be bigger than %d in any dimension.\n", 255);
        exit(1);
    }

    if (fillFactor < 0)
    {
        fprintf(stderr, "Error: The fill factor cannot be negative.\n");
        exit(1);
    }
    else if (fillFactor > 255)
    {
        fprintf(stderr, "Error: The fill factor cannot exceed 255.\n");
    }

    if (trials < 1)
    {
        fprintf(stderr, "Error: You must allow at least one trial.\n");
        exit(1);
    }
}

/** Seed the PRNG with some data from /dev/urandom.
 *
 * We will read however many bytes are required for an unsigned int,
 * and call srand().
 */
void seed_rand(void)
{
    FILE *f;
    unsigned int seed = 0;

    if (NULL != (f = fopen("/dev/urandom", "rb+")))
    {
        for (int i = 0; i < (sizeof(unsigned int) / sizeof(char)); ++i)
        {
            unsigned char c;

            if (1 == fread(&c, 1, 1, f))
            {
                seed <<= 8;
                seed |= c;
            } 
        }

        fclose(f);
    }

    srand(seed);
}


/** Main program entry point
 *
 * Read the options from the command line.
 * Seed the PRNG.
 * Run a given number f trials.
 * Collate the results.
 * Print the results.
 */
int main(int argc, char **argv)
{
    std::list<Life> maps();
    std::map<unsigned int, unsigned int, ltint> histogram;

    readopts(argc, argv);

    seed_rand();

    for (int trial = 0; trial < trials; ++trial)
    {
        Life *board = new Life(sizeX, sizeY, historySize);
        board->randomiseBoard(fillFactor);

        unsigned int iter = 0;

        while (board->play() && (iter < maxIterations))
        {
            ++iter;
        }

        if (histogram.count(iter) == 0)
        {
            histogram[iter] = 1;
        }
        else
        {   
            histogram[iter] += 1;
        }

        delete board;
        
    }

    /* Now process the results */
    std::map<unsigned int, unsigned int, ltint>::iterator cur = histogram.begin();

    std::cout << std::setw(8) << "# Iter" << "   " << "Frequency" << std::endl;
    for (cur = histogram.begin(); cur != histogram.end(); ++cur)
    {
        if (cur->second != 0)
            std::cout << std::setw(8) << cur->first << "   " << cur->second << std::endl;
    }

    return 0;
}
